# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.
# 
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from os.path import expanduser, dirname, join
from glob import glob
from itertools import chain
from subprocess import check_output, CalledProcessError
import sys
import distutils.unixccompiler

__version__ = '1.4.0'

###############################################################################
# Monkey-patch setuptools to compile in parallel (copied from pytorch)
###############################################################################
original_link = distutils.unixccompiler.UnixCCompiler.link


def parallelCCompile(self, sources, output_dir=None, macros=None,
                     include_dirs=None, debug=0, extra_preargs=None,
                     extra_postargs=None, depends=None):
    # those lines are copied from distutils.ccompiler.CCompiler directly
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs)
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    # compile using a thread pool
    import multiprocessing.pool

    def _single_compile(obj):
        src, ext = build[obj]
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
    num_jobs = multiprocessing.cpu_count()
    multiprocessing.pool.ThreadPool(num_jobs).map(_single_compile, objects)

    return objects


def patched_link(self, *args, **kwargs):
    _cxx = self.compiler_cxx
    self.compiler_cxx = None
    result = original_link(self, *args, **kwargs)
    self.compiler_cxx = _cxx
    return result


distutils.ccompiler.CCompiler.compile = parallelCCompile
distutils.unixccompiler.UnixCCompiler.link = patched_link


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


sources = list(chain(
    glob('py/*.cpp'),
    glob('replayer/*.cpp'),
    glob('client/*.cpp'),
))
print(sources)

ext_modules = [
    Extension(
        'torchcraft',
        sources,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            "include/torchcraft",
            "include",
            "replayer",
            ".",
            "BWEnv/fbs",
        ],
        # TODO Search for ZSTD and define this if it exists
        define_macros=[('WITH_ZSTD', None)],
        libraries=['zstd', 'zmq'],
        language='c++'
    ),
]


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append("-std=c++11")
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


setup(
    name='torchcraft',
    version=__version__,
    author='Zeming Lin',
    author_email='zlin@fb.com',
    url='',
    description='Torchcraft',
    long_description='',
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.1'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
)



import numpy
from typing import *
from .. import tc2lib
from tc2 import CommandType
from tc2.constants import *


def command_attack_unit(unit: tc2lib.Unit, target: tc2lib.Unit) -> CommandType:
    return [
        int(tc2lib.Constants.command_unit),
        int(unit.id),
        int(tc2lib.Constants.unitcommandtypes.Attack_Unit),
        int(target.id)]


def command_attack_position(unit: tc2lib.Unit, x: int, y: int) -> CommandType:
    return [
        int(tc2lib.Constants.command_unit),
        int(unit.id),
        int(tc2lib.Constants.unitcommandtypes.Attack_Move),
        int(-1),
        int(x),
        int(y)]


def command_move(unit: tc2lib.Unit, x: int, y: int) -> CommandType:
    return [
        int(tc2lib.Constants.command_unit),
        int(unit.id),
        int(tc2lib.Constants.unitcommandtypes.Move),
        int(-1),
        int(x),
        int(y)]


def command_hold_position(unit: tc2lib.Unit) -> CommandType:
    return [
        int(tc2lib.Constants.command_unit),
        int(unit.id),
        int(tc2lib.Constants.unitcommandtypes.Hold_Position)]


def command_gather(unit: tc2lib.Unit, target: tc2lib.Unit) -> CommandType:
    return [
        int(tc2lib.Constants.command_unit),
        int(unit.id),
        int(tc2lib.Constants.unitcommandtypes.Gather),
        int(target.id)]


def command_draw_unit_circle(unit: tc2lib.Unit, radius: int, color: int = 255) -> CommandType:
    return [
        int(TCCommands.draw_unit_circle),
        int(unit.id),
        int(radius),
        int(color)]


class CommandTensorShapeError(ValueError):
    pass


class CommandTensorUnitDimensionError(ValueError):
    pass


class CommandTensorCommandDimensionError(ValueError):
    pass


non_type_columns = 3  # Target ID + Target X + Target Y


def commandtype_columns() -> int:
    """The number of columns used to represent TCUnitCommands"""
    return len(TCUnitCommands.ids)


def buildtype_columns() -> int:
    """The number of columns used to store possible Unit/Tech/Upgrade types"""
    return len(TCUnits.ids) + len(TCTechs.ids) + len(TCUpgrades.ids)


def command_tensor_shape(observation) -> Tuple[int, int]:
    """Returns the expected shape of a command tensor for a given observation"""
    return (len(observation.units.agent.ids), non_type_columns + commandtype_columns() + buildtype_columns())


def tensor_to_commands(observation, command_tensor: numpy.ndarray) -> List[CommandType]:
    """Converts a Units x [One-hot command, one-hot type ID, Target ID, Target X, Target Y] tensor to TorchCraft commands"""

    # Verify appropriate dimensions
    expected_shape = command_tensor_shape(observation)

    if len(command_tensor.shape) != 2:
        raise CommandTensorShapeError(
            "The command tensor should be two dimensional: [# of units in the unit features] x [features]. Should be {} but was {}"
            .format(expected_shape, command_tensor.shape))
    if command_tensor.shape[0] != expected_shape[0]:
        raise CommandTensorUnitDimensionError(
            "The length of the first dimension of the command tensor should equal the number of units you have. Should be {} but was {}"
            .format(expected_shape[0], command_tensor.shape[0]))
    if command_tensor.shape[1] != expected_shape[1]:
        raise CommandTensorCommandDimensionError(
            "The length of the second dimension of the command tensor should equal the number of features. Should be {} but was {}"
            .format(expected_shape[1], command_tensor.shape[1]))

    # Convert the tensor to commands
    commands_output: List[CommandType] = []
    for unit_index in range(0, expected_shape[0]):
        unit_id = observation.units.agent.ids[unit_index]
        unit_tensor = command_tensor[unit_index]
        target_id = unit_tensor[0]
        target_x = unit_tensor[1]
        target_y = unit_tensor[2]
        command_id = numpy.argmax(unit_tensor[non_type_columns: non_type_columns + commandtype_columns()])
        build_id = numpy.argmax(unit_tensor[non_type_columns + commandtype_columns():])
        if build_id >= len(TCUnits.ids):
            # It's not a unit type; it's a tech or upgrade type
            build_id -= len(TCUnits.ids)
            if build_id >= len(TCTechs.ids):
                # It's not a tech type; it's an upgrade type
                build_id -= len(TCTechs.ids)
        commands_output += [[
            int(tc2lib.Constants.command_unit),
            int(unit_id),
            int(command_id),
            # int(build_id),
            int(target_id),
            int(target_x),
            int(target_y)]]
    return commands_output



#!/usr/bin/env python3
# #############################################################################
# Copyright (c) 2018-present    lzutao <taolzu(at)gmail.com>
# All rights reserved.
#
# This source code is licensed under both the BSD-style license (found in the
# LICENSE file in the root directory of this source tree) and the GPLv2 (found
# in the COPYING file in the root directory of this source tree).
# #############################################################################
import os
import subprocess
import tempfile


def valgrindTest(valgrind, datagen, fuzzer, zstd, fullbench):
  VALGRIND_ARGS = [valgrind, '--leak-check=full', '--show-leak-kinds=all', '--error-exitcode=1']

  print('\n ---- valgrind tests : memory analyzer ----')

  subprocess.check_call([*VALGRIND_ARGS, datagen, '-g50M'], stdout=subprocess.DEVNULL)

  if subprocess.call([*VALGRIND_ARGS, zstd],
                     stdout=subprocess.DEVNULL) == 0:
    raise subprocess.CalledProcessError('zstd without argument should have failed')

  with subprocess.Popen([datagen, '-g80'], stdout=subprocess.PIPE) as p1, \
       subprocess.Popen([*VALGRIND_ARGS, zstd, '-', '-c'],
                        stdin=p1.stdout,
                        stdout=subprocess.DEVNULL) as p2:
    p1.stdout.close()  # Allow p1 to receive a SIGPIPE if p2 exits.
    p2.communicate()
    if p2.returncode != 0:
      raise subprocess.CalledProcessError()

  with subprocess.Popen([datagen, '-g16KB'], stdout=subprocess.PIPE) as p1, \
       subprocess.Popen([*VALGRIND_ARGS, zstd, '-vf', '-', '-c'],
                        stdin=p1.stdout,
                        stdout=subprocess.DEVNULL) as p2:
    p1.stdout.close()
    p2.communicate()
    if p2.returncode != 0:
      raise subprocess.CalledProcessError()

  with tempfile.NamedTemporaryFile() as tmp_fd:
    with subprocess.Popen([datagen, '-g2930KB'], stdout=subprocess.PIPE) as p1, \
         subprocess.Popen([*VALGRIND_ARGS, zstd, '-5', '-vf', '-', '-o', tmp_fd.name],
                          stdin=p1.stdout) as p2:
      p1.stdout.close()
      p2.communicate()
      if p2.returncode != 0:
        raise subprocess.CalledProcessError()

    subprocess.check_call([*VALGRIND_ARGS, zstd, '-vdf', tmp_fd.name, '-c'],
                          stdout=subprocess.DEVNULL)

    with subprocess.Popen([datagen, '-g64MB'], stdout=subprocess.PIPE) as p1, \
         subprocess.Popen([*VALGRIND_ARGS, zstd, '-vf', '-', '-c'],
                          stdin=p1.stdout,
                          stdout=subprocess.DEVNULL) as p2:
      p1.stdout.close()
      p2.communicate()
      if p2.returncode != 0:
        raise subprocess.CalledProcessError()

  subprocess.check_call([*VALGRIND_ARGS, fuzzer, '-T1mn', '-t1'])
  subprocess.check_call([*VALGRIND_ARGS, fullbench, '-i1'])


def main():
  import argparse
  parser = argparse.ArgumentParser(description='Valgrind tests : memory analyzer')
  parser.add_argument('valgrind', help='valgrind path')
  parser.add_argument('zstd', help='zstd path')
  parser.add_argument('datagen', help='datagen path')
  parser.add_argument('fuzzer', help='fuzzer path')
  parser.add_argument('fullbench', help='fullbench path')

  args = parser.parse_args()

  valgrind = args.valgrind
  zstd = args.zstd
  datagen = args.datagen
  fuzzer = args.fuzzer
  fullbench = args.fullbench

  valgrindTest(valgrind, datagen, fuzzer, zstd, fullbench)


if __name__ == '__main__':
  main()



from tc2 import tc2lib, CommandType


def command_spawn(player: int, unit: int, x: int, y: int) -> CommandType:
    return [
        int(tc2lib.Constants.command_openbw),
        int(tc2lib.Constants.openbwcommandtypes.SpawnUnit),
        int(player),
        int(unit),
        int(x),
        int(y)]


def command_kill(unit_id: int) -> CommandType:
    return [
        int(tc2lib.Constants.command_openbw),
        int(tc2lib.Constants.openbwcommandtypes.KillUnit),
        int(unit_id)]


def command_upgrade(player: int, upgrade: int, level: int = 1) -> CommandType:
    return [
        int(tc2lib.Constants.command_openbw),
        int(tc2lib.Constants.openbwcommandtypes.SetPlayerUpgradeLevel),
        int(player),
        int(upgrade),
        int(level)]


def command_research(player: int, tech: int, researched: bool = True) -> CommandType:
    return [
        int(tc2lib.Constants.command_openbw),
        int(tc2lib.Constants.openbwcommandtypes.SetPlayerResearched),
        int(player),
        int(tech),
        int(researched)]



from tc2.constants import *
from tc2.envs import TC2Env
from tc2.rules import *
from random import randint, sample
from typing import List, Tuple
import itertools


class TC2EnvMoveToBeaconReward(TC2RewardAbstract):
    def __init__(self):
        self._wins = 0
        super().__init__()

    def add_win(self) -> None:
        self._wins += 1

    def reward(self, tcstates) -> List[float]:
        return [self._wins, 0]

    def done(self, tcstates) -> bool:
        # PySC2 MoveToBeacon is defined at 120 seconds but 16 FPS, unlike the 24 FPS standard for SC1
        return tcstates[0].frame_from_bwapi >= 16 * 120


class TC2EnvMoveToBeacon(TC2Env):
    """
    Reproduction of the MoveToBeacon scenario from PySC2
    See https://github.com/deepmind/pysc2/blob/master/docs/mini_games.md
    """

    def __init__(self) -> None:
        super().__init__()
        self.walker_type = TCUnits.Terran_Civilian
        self.beacon_type = TCUnits.Zerg_Overlord
        self._reward: TC2EnvMoveToBeaconReward

    def _before_reset(self) -> None:
        self.set_map('maps/empty64.scm')
        self.set_complete_observability(True)
        self.set_reward(TC2EnvMoveToBeaconReward())

    def _random_pixel(self):
        return randint(32 * 3, 32 * 61)

    def _within_distance(self, a: Tuple[int, int], b: Tuple[int, int], distance_max: int) -> bool:
        distance_squared = (a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2
        distance_max_squared = distance_max ** 2
        return distance_squared <= distance_max_squared

    def _reset_beacon(self) -> None:
        units = list(itertools.chain.from_iterable(self.get_raw_state().units.values()))
        for unit in units:
            self._kill(unit.id)
        a = None
        b = None
        # Note that this distance check is in pixels
        while a is None or self._within_distance(a, b, 32 * 4):
            a = (self._random_pixel(), self._random_pixel())
            b = (self._random_pixel(), self._random_pixel())
        self._spawn(TCAgent, self.walker_type, a[0], a[1])
        self._spawn(TCEnemy, self.beacon_type, b[0], b[1])
        self.skip_frames_until_commands_executed()

    def _after_step(self) -> None:
        units = list(itertools.chain.from_iterable(self.get_raw_state().units.values()))
        walker = next((unit for unit in units if unit.type == self.walker_type), None)
        beacon = next((unit for unit in units if unit.type == self.beacon_type), None)
        # Note that this distance check is in walktiles (8x8 pixels)
        if walker is not None and beacon is not None and self._within_distance((walker.x, walker.y), (beacon.x, beacon.y), 8 * 2):
            self._reward.add_win()
            self._reset_beacon()

    def _after_reset(self) -> None:
        self._reset_beacon()



from typing import List, Tuple
from tc2 import CommandType, tc2lib
from tc2.tools import *
from tc2.constants import *


def do_nothing(tcstate: tc2lib.State) -> List[CommandType]:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    return []


def our_units(tcstate: tc2lib.State) -> List[tc2lib.Unit]:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    player = tcstate.player_id
    units = tcstate.units
    return units[player] if player in units else []


def enemy_units(tcstate: tc2lib.State) -> List:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    player_id = other_player_id(tcstate.player_id)
    units = tcstate.units
    return units[player_id] if player_id in units else []


def neutral_units(tcstate: tc2lib.State) -> List:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    return tcstate.units[TCNeutral.player_id(tcstate)]


def hold_position(tcstate: tc2lib.State) -> List[CommandType]:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    units_ours = our_units(tcstate)
    return [
        command_hold_position(unit)
        for unit in units_ours
        if not unit.holding_position]


def attack_middle(tcstate: tc2lib.State) -> List[CommandType]:
    if not isinstance(tcstate, tc2lib.State):
        raise TypeError(f"tcstate argument should be of type tc2lib.State, got {tcstate}")
    middle_x = tcstate.map_size[0] / 2
    middle_y = tcstate.map_size[1] / 2
    units_ours = our_units(tcstate)
    return [
        command_attack_position(unit, middle_x, middle_y)
        for unit in units_ours
        if unit.idle]


def distance_squared(a: Tuple[float, float], b: Tuple[float, float]) -> float:
    return (a[0] - b[0]) * (a[0] - b[0]) + (a[1] - b[1]) * (a[1] - b[1])


def unit_distance_center_squared(a, b) -> float:
    return distance_squared((a.x, a.y), (b.x, b.y))


def can_attack(attacker, target) -> bool:
    if (attacker.type in [TCUnits.Terran_Bunker, TCUnits.Protoss_Carrier]):
        return True
    if (attacker.type == TCUnits.Protoss_Reaver and not target.flying):
        return True
    return attacker.airATK > 0 if target.flying else attacker.groundATK > 0


def attack_closest_enemy(attacker, enemies: List) -> List[CommandType]:
    valid_targets = list(filter(lambda target: can_attack(attacker, target), enemies))
    if valid_targets:
        # Attack the closest valid target
        target = min(valid_targets, key=lambda target: unit_distance_center_squared(attacker, target))
        return [command_attack_unit(attacker, target)] if attacker.command.targetId != target.id else []
    elif enemies and attacker.idle:
        # Attack the position of the enemy closest to the centroid of enemies
        centroid = (
            sum(map(lambda enemy: enemy.x, enemies)) / len(enemies),
            sum(map(lambda enemy: enemy.y, enemies)) / len(enemies))
        target_unit = min(enemies, key=lambda target: distance_squared(centroid, (target.y, target.y)))
        return [command_attack_position(attacker, target_unit.x, target_unit.y)]
    else:
        # Do nothing
        return []


def gather_closest_resource(gatherer, resources: List) -> List[CommandType]:
    if resources:
        resource = min(resources, key=lambda target: unit_distance_center_squared(gatherer, target))
        return [command_gather(gatherer, resource)]
    else:
        return []


def attack_closest(tcstate) -> List[CommandType]:
    units_ours = our_units(tcstate)
    units_enemy = enemy_units(tcstate)
    commands = [attack_closest_enemy(unit, units_enemy) for unit in units_ours]
    output = [item for sublist in commands for item in sublist]
    return output


def gather_closest(tcstate) -> List[CommandType]:
    units_ours = our_units(tcstate)
    resources = list(filter(
        lambda unit: unit.visible and unit.type in [
            TCUnits.Resource_Mineral_Field,
            TCUnits.Terran_Refinery,
            TCUnits.Protoss_Assimilator,
            TCUnits.Zerg_Extractor],
        enemy_units(tcstate) + neutral_units(tcstate)))
    commands = [
        gather_closest_resource(unit, resources)
        for unit in units_ours
        if unit.idle]
    output = [item for sublist in commands for item in sublist]
    return output



import os
import json
import zstandard as zstd
import logging
import numpy as np
from typing import NamedTuple, List, Dict, Any, NewType, Optional, Callable, Union, Tuple
from collections import defaultdict
from tc2 import CommandType, FrameNum, UnitId
from tc2.constants import TCUnits, TCCommands


class CVisDrawCommand(NamedTuple):
    code: int
    args: List[int]
    str: str  # type: ignore
    cherrypi_ids_args_indices: List[int]


class CVisLog(NamedTuple):
    frame: int
    attachments: List[Any]
    file: str
    line: int
    message: str
    sev: int = 0  # Severity


class CVisUnitData(NamedTuple):
    last_seen_type: int = -1


class CVisUnitFirstSeen(NamedTuple):
    id: int
    type: int
    x: int  # pixel_x
    y: int  # pixel_y


def get_tensor_summary(name: str, tensor: np.ndarray) -> Dict[str, Any]:
    if tensor.size == 0:
        raise Exception(f"Cannot produce a summary for empty tensor {name}")
    HIST_NUM_BUCKETS = 10
    hist_count, bin_edges = np.histogram(tensor, HIST_NUM_BUCKETS)
    return {
        "shape": list(tensor.shape),
        "type": tensor.dtype.name,
        "min": tensor.min().item(),
        "max": tensor.max().item(),
        "mean": tensor.mean().item(),
        "std": tensor.std().item(),
        "median": np.median(tensor).item(),
        "absmedian": np.median(np.abs(tensor)).item(),
        "name": name,
        "hist": {
            "num_buckets": HIST_NUM_BUCKETS,
            "min": bin_edges[0].item(),
            "max": bin_edges[-1].item(),
            "values": [v.item() for v in hist_count],
        },
    }


# pylint: disable=too-many-instance-attributes
class CVisDumper:
    def __init__(self, replay_file: str, suffix: str = ''):
        self._current_frame: FrameNum = FrameNum(0)
        self._replay_file: str = self.parse_replay_file(replay_file)
        self._suffix: str = '' if suffix == '' else f'.{suffix}'

        # Trace
        self._tasks: List[Any] = []  # Not implemented
        self._board_updates: List[Any] = []  # Not implemented
        self._draw_commands: Dict[int, List[CVisDrawCommand]] = defaultdict(
            list)
        self._game_values: Dict[FrameNum, Dict[str, float]] = defaultdict(dict)

        # Units
        self._units_info: Dict[UnitId, CVisUnitData] = defaultdict(
            CVisUnitData)
        self._units_updates: Dict[UnitId, Dict[FrameNum, dict]] = defaultdict(
            dict)
        self._units_first_seen: Dict[FrameNum, List[
            CVisUnitFirstSeen]] = defaultdict(list)

        # Logs
        self._logs: List[CVisLog] = []
        self._units_logs: Dict[int, List[CVisLog]] = defaultdict(list)

        self._cleanup_fn: List[Callable[[], None]] = []

        # Tensors
        self._tensor_summaries: Dict[FrameNum, List[dict]] = defaultdict(list)
        self._heatmaps_metadata: List[dict] = []
        self._heatmaps_writers: Dict[str, Any] = {}

    def step(self, tcstate) -> None:
        """
        Called right after we reach a new frame
        """
        frame = FrameNum(tcstate.frame_from_bwapi)
        self._current_frame = frame
        for listUnits in tcstate.units.values():
            for u in listUnits:
                updated = False
                if u.id not in self._units_info:
                    # New unit
                    self._units_info[u.id] = CVisUnitData(u.type)
                    self._units_first_seen[frame].append(
                        CVisUnitFirstSeen(
                            id=u.id, type=u.type, x=u.pixel_x, y=u.pixel_y))
                    updated = True
                elif u.type != self._units_info[u.id].last_seen_type:
                    updated = True
                if updated:
                    self._units_updates[u.id][frame] = {'type': u.type}

    def on_send(self, cmds: List[CommandType]) -> None:
        drawCmdToUnitIds: Dict[int, List[int]] = {
            TCCommands.draw_line: [],
            TCCommands.draw_unit_line: [0, 1],
            TCCommands.draw_unit_pos_line: [0],
            TCCommands.draw_unit_circle: [0],
            TCCommands.draw_circle: [],
            TCCommands.draw_text: [],
            TCCommands.draw_text_screen: [],
        }
        for cmd in cmds:
            if len(cmd) < 2 or cmd[0] not in drawCmdToUnitIds:
                continue
            self._draw_commands[self._current_frame].append(
                CVisDrawCommand(
                    code=cmd[0],
                    args=cmd[2:] if isinstance(cmd[1], str) else cmd[1:],
                    str=cmd[1] if isinstance(cmd[1], str) else '',
                    cherrypi_ids_args_indices=drawCmdToUnitIds[cmd[0]],
                ))

    def add_logger(self, logger: logging.Logger, unit_id: int,
                   level: Union[str, int]) -> None:
        class CVisLogHandler(logging.Handler):
            def __init__(self, cvis: 'CVisDumper', logs_list: List[CVisLog]):
                self.cvis = cvis
                self.logs_list = logs_list
                super().__init__()

            def emit(self, record: logging.LogRecord) -> None:
                self.logs_list.append(
                    CVisLog(
                        frame=self.cvis._current_frame,
                        attachments=[],
                        file=record.filename,
                        line=record.lineno,
                        message=record.getMessage(),
                        sev=record.levelno,
                    ))

        hdlr = CVisLogHandler(self, self._logs
                              if unit_id < 0 else self._units_logs[unit_id])
        hdlr.setLevel(level)
        logger.addHandler(hdlr)
        self._cleanup_fn.append(lambda: logger.removeHandler(hdlr))

    def dump_game_value(self, key: str, val: float,
                        frame: FrameNum = -1) -> None:
        if frame < 0:
            frame = self._current_frame
        self._game_values[frame][key] = val

    def write_trace(self) -> None:
        if self._replay_file == '':
            return

        def dump_list(l: list) -> list:
            return [x._asdict() for x in l]

        build_types_to_names: Dict[int, str] = {}
        for unit_name, unit_id in TCUnits.names.items():
            build_types_to_names[int(unit_id)] = unit_name

        trace = {
            "types_names": build_types_to_names,
            "tasks": self._tasks,
            "logs": dump_list(self._logs),
            "units_logs":
            {unit: dump_list(logs)
             for unit, logs in self._units_logs.items()},
            "units_updates": self._units_updates,
            "units_first_seen":
            {unit: dump_list(l)
             for unit, l in self._units_first_seen.items()},
            "board_updates": self._board_updates,
            "draw_commands":
            {frame: dump_list(c)
             for frame, c in self._draw_commands.items()},
            "trees": [],
            "heatmaps": self._heatmaps_metadata,
            "tensors_summaries": self._tensor_summaries,
            "game_values": self._game_values,
            "_version": 0
        }
        folder = f'{self._replay_file}.cvis{self._suffix}'

        # Cleanup first, because the IO might fail
        self._replay_file = ''
        for f in self._cleanup_fn:
            f()
        for w in self._heatmaps_writers.values():
            w.close()
        self._heatmaps_writers = {}
        self._cleanup_fn = []

        os.makedirs(folder, exist_ok=True)
        with open(os.path.join(folder, 'trace.json'), 'wb+') as fh:
            cctx = zstd.ZstdCompressor()
            with cctx.stream_writer(fh) as compressor:
                compressor.write(json.dumps(trace).encode('utf-8'))

    def parse_replay_file(self, file: str) -> str:
        # Function taken from OpenBW
        # See https://github.com/OpenBW/bwapi/blob/develop-openbw/bwapi/BWAPI/Source/BWAPI/GameEvents.cpp#L380
        def replace(cstr: str) -> str:
            if cstr == '\\':
                return '/'
            c = ord(cstr)
            if c >= 128:
                return cstr
            if ord('a') <= c <= ord('z') or ord('A') <= c <= ord('Z') or ord('0') <= c <= ord('9'):
                return cstr
            if cstr in ['-', '-', '_', '_', '.', '/', ' ']:
                return cstr
            return ''

        return ''.join([replace(c) for c in file])

    def dump_tensor_summary(self, name: str, tensor: np.ndarray) -> None:
        if tensor.size == 0:
            raise Exception(f"Empty tensor {name}")
        self._tensor_summaries[self._current_frame].append(
            get_tensor_summary(name, tensor))

    def dump_terrain_heatmap(
            self,
            name: str,
            tensor: np.ndarray,
            top_left_pixel: Tuple[int, int] = (0, 0),
            scaling_to_pixels: Tuple[float, float] = (8, 8)) -> None:
        if len(tensor.shape) != 2:
            raise Exception(
                f"Tensor {name} should be 2d, but has shape {tensor.shape}")
        if tensor.size == 0:
            raise Exception(f"Empty tensor {name}")

        if name not in self._heatmaps_writers:
            dump_directory = f'{self._replay_file}.cvis{self._suffix}'
            os.makedirs(dump_directory, exist_ok=True)
            filename = f'{dump_directory}/tensor__{len(self._heatmaps_writers)}__f{self._current_frame}.json.zstd.stream'

            # Open file to write tensor values in streaming
            fh = open(filename, 'wb+')
            cctx = zstd.ZstdCompressor()
            self._heatmaps_writers[name] = cctx.stream_writer(fh)

            # Write initial metadata
            self._heatmaps_metadata.append({
                "name": name,
                "filename": filename,
                "first_frame": self._current_frame,
            })

        summary = get_tensor_summary(name, tensor)
        cvis_tensor = {
            "top_left_pixel": top_left_pixel,
            "scaling": scaling_to_pixels,
            "dimension": tensor.shape,
            "data": tensor.flatten('C').tolist(),
            "summary": summary,
        }
        self._heatmaps_writers[name].write(
            json.dumps({
                "key": self._current_frame,
                "value": cvis_tensor,
            }).encode('utf-8'))



#!/usr/bin/env python

# ################################################################
# Copyright (c) 2016-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under both the BSD-style license (found in the
# LICENSE file in the root directory of this source tree) and the GPLv2 (found
# in the COPYING file in the root directory of this source tree).
# ##########################################################################

import argparse
import contextlib
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile


def abs_join(a, *p):
    return os.path.abspath(os.path.join(a, *p))


# Constants
FUZZ_DIR = os.path.abspath(os.path.dirname(__file__))
CORPORA_DIR = abs_join(FUZZ_DIR, 'corpora')
TARGETS = [
    'simple_round_trip',
    'stream_round_trip',
    'block_round_trip',
    'simple_decompress',
    'stream_decompress',
    'block_decompress',
    'dictionary_round_trip',
    'dictionary_decompress',
    'zstd_frame_info',
]
ALL_TARGETS = TARGETS + ['all']
FUZZ_RNG_SEED_SIZE = 4

# Standard environment variables
CC = os.environ.get('CC', 'cc')
CXX = os.environ.get('CXX', 'c++')
CPPFLAGS = os.environ.get('CPPFLAGS', '')
CFLAGS = os.environ.get('CFLAGS', '-O3')
CXXFLAGS = os.environ.get('CXXFLAGS', CFLAGS)
LDFLAGS = os.environ.get('LDFLAGS', '')
MFLAGS = os.environ.get('MFLAGS', '-j')

# Fuzzing environment variables
LIB_FUZZING_ENGINE = os.environ.get('LIB_FUZZING_ENGINE', 'libregression.a')
AFL_FUZZ = os.environ.get('AFL_FUZZ', 'afl-fuzz')
DECODECORPUS = os.environ.get('DECODECORPUS',
                              abs_join(FUZZ_DIR, '..', 'decodecorpus'))

# Sanitizer environment variables
MSAN_EXTRA_CPPFLAGS = os.environ.get('MSAN_EXTRA_CPPFLAGS', '')
MSAN_EXTRA_CFLAGS = os.environ.get('MSAN_EXTRA_CFLAGS', '')
MSAN_EXTRA_CXXFLAGS = os.environ.get('MSAN_EXTRA_CXXFLAGS', '')
MSAN_EXTRA_LDFLAGS = os.environ.get('MSAN_EXTRA_LDFLAGS', '')


def create(r):
    d = os.path.abspath(r)
    if not os.path.isdir(d):
        os.mkdir(d)
    return d


def check(r):
    d = os.path.abspath(r)
    if not os.path.isdir(d):
        return None
    return d


@contextlib.contextmanager
def tmpdir():
    dirpath = tempfile.mkdtemp()
    try:
        yield dirpath
    finally:
        shutil.rmtree(dirpath, ignore_errors=True)


def parse_targets(in_targets):
    targets = set()
    for target in in_targets:
        if not target:
            continue
        if target == 'all':
            targets = targets.union(TARGETS)
        elif target in TARGETS:
            targets.add(target)
        else:
            raise RuntimeError('{} is not a valid target'.format(target))
    return list(targets)


def targets_parser(args, description):
    parser = argparse.ArgumentParser(prog=args.pop(0), description=description)
    parser.add_argument(
        'TARGET',
        nargs='*',
        type=str,
        help='Fuzz target(s) to build {{{}}}'.format(', '.join(ALL_TARGETS)))
    args, extra = parser.parse_known_args(args)
    args.extra = extra

    args.TARGET = parse_targets(args.TARGET)

    return args


def parse_env_flags(args, flags):
    """
    Look for flags set by environment variables.
    """
    san_flags = ','.join(re.findall('-fsanitize=((?:[a-z]+,?)+)', flags))
    nosan_flags = ','.join(re.findall('-fno-sanitize=((?:[a-z]+,?)+)', flags))

    def set_sanitizer(sanitizer, default, san, nosan):
        if sanitizer in san and sanitizer in nosan:
            raise RuntimeError('-fno-sanitize={s} and -fsanitize={s} passed'.
                               format(s=sanitizer))
        if sanitizer in san:
            return True
        if sanitizer in nosan:
            return False
        return default

    san = set(san_flags.split(','))
    nosan = set(nosan_flags.split(','))

    args.asan = set_sanitizer('address', args.asan, san, nosan)
    args.msan = set_sanitizer('memory', args.msan, san, nosan)
    args.ubsan = set_sanitizer('undefined', args.ubsan, san, nosan)

    args.sanitize = args.asan or args.msan or args.ubsan

    return args


def compiler_version(cc, cxx):
    """
    Determines the compiler and version.
    Only works for clang and gcc.
    """
    cc_version_bytes = subprocess.check_output([cc, "--version"])
    cxx_version_bytes = subprocess.check_output([cxx, "--version"])
    compiler = None
    version = None
    if b'clang' in cc_version_bytes:
        assert(b'clang' in cxx_version_bytes)
        compiler = 'clang'
    elif b'gcc' in cc_version_bytes:
        assert(b'gcc' in cxx_version_bytes)
        compiler = 'gcc'
    if compiler is not None:
        version_regex = b'([0-9])+\.([0-9])+\.([0-9])+'
        version_match = re.search(version_regex, cc_version_bytes)
        version = tuple(int(version_match.group(i)) for i in range(1, 4))
    return compiler, version


def overflow_ubsan_flags(cc, cxx):
    compiler, version = compiler_version(cc, cxx)
    if compiler == 'gcc':
        return ['-fno-sanitize=signed-integer-overflow']
    if compiler == 'clang' and version >= (5, 0, 0):
        return ['-fno-sanitize=pointer-overflow']
    return []


def build_parser(args):
    description = """
    Cleans the repository and builds a fuzz target (or all).
    Many flags default to environment variables (default says $X='y').
    Options that aren't enabling features default to the correct values for
    zstd.
    Enable sanitizers with --enable-*san.
    For regression testing just build.
    For libFuzzer set LIB_FUZZING_ENGINE and pass --enable-coverage.
    For AFL set CC and CXX to AFL's compilers and set
    LIB_FUZZING_ENGINE='libregression.a'.
    """
    parser = argparse.ArgumentParser(prog=args.pop(0), description=description)
    parser.add_argument(
        '--lib-fuzzing-engine',
        dest='lib_fuzzing_engine',
        type=str,
        default=LIB_FUZZING_ENGINE,
        help=('The fuzzing engine to use e.g. /path/to/libFuzzer.a '
              "(default: $LIB_FUZZING_ENGINE='{})".format(LIB_FUZZING_ENGINE)))

    fuzz_group = parser.add_mutually_exclusive_group()
    fuzz_group.add_argument(
        '--enable-coverage',
        dest='coverage',
        action='store_true',
        help='Enable coverage instrumentation (-fsanitize-coverage)')
    fuzz_group.add_argument(
        '--enable-fuzzer',
        dest='fuzzer',
        action='store_true',
        help=('Enable clang fuzzer (-fsanitize=fuzzer). When enabled '
              'LIB_FUZZING_ENGINE is ignored')
    )

    parser.add_argument(
        '--enable-asan', dest='asan', action='store_true', help='Enable UBSAN')
    parser.add_argument(
        '--enable-ubsan',
        dest='ubsan',
        action='store_true',
        help='Enable UBSAN')
    parser.add_argument(
        '--enable-ubsan-pointer-overflow',
        dest='ubsan_pointer_overflow',
        action='store_true',
        help='Enable UBSAN pointer overflow check (known failure)')
    parser.add_argument(
        '--enable-msan', dest='msan', action='store_true', help='Enable MSAN')
    parser.add_argument(
        '--enable-msan-track-origins', dest='msan_track_origins',
        action='store_true', help='Enable MSAN origin tracking')
    parser.add_argument(
        '--msan-extra-cppflags',
        dest='msan_extra_cppflags',
        type=str,
        default=MSAN_EXTRA_CPPFLAGS,
        help="Extra CPPFLAGS for MSAN (default: $MSAN_EXTRA_CPPFLAGS='{}')".
        format(MSAN_EXTRA_CPPFLAGS))
    parser.add_argument(
        '--msan-extra-cflags',
        dest='msan_extra_cflags',
        type=str,
        default=MSAN_EXTRA_CFLAGS,
        help="Extra CFLAGS for MSAN (default: $MSAN_EXTRA_CFLAGS='{}')".format(
            MSAN_EXTRA_CFLAGS))
    parser.add_argument(
        '--msan-extra-cxxflags',
        dest='msan_extra_cxxflags',
        type=str,
        default=MSAN_EXTRA_CXXFLAGS,
        help="Extra CXXFLAGS for MSAN (default: $MSAN_EXTRA_CXXFLAGS='{}')".
        format(MSAN_EXTRA_CXXFLAGS))
    parser.add_argument(
        '--msan-extra-ldflags',
        dest='msan_extra_ldflags',
        type=str,
        default=MSAN_EXTRA_LDFLAGS,
        help="Extra LDFLAGS for MSAN (default: $MSAN_EXTRA_LDFLAGS='{}')".
        format(MSAN_EXTRA_LDFLAGS))
    parser.add_argument(
        '--enable-sanitize-recover',
        dest='sanitize_recover',
        action='store_true',
        help='Non-fatal sanitizer errors where possible')
    parser.add_argument(
        '--debug',
        dest='debug',
        type=int,
        default=1,
        help='Set DEBUGLEVEL (default: 1)')
    parser.add_argument(
        '--force-memory-access',
        dest='memory_access',
        type=int,
        default=0,
        help='Set MEM_FORCE_MEMORY_ACCESS (default: 0)')
    parser.add_argument(
        '--fuzz-rng-seed-size',
        dest='fuzz_rng_seed_size',
        type=int,
        default=4,
        help='Set FUZZ_RNG_SEED_SIZE (default: 4)')
    parser.add_argument(
        '--disable-fuzzing-mode',
        dest='fuzzing_mode',
        action='store_false',
        help='Do not define FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION')
    parser.add_argument(
        '--enable-stateful-fuzzing',
        dest='stateful_fuzzing',
        action='store_true',
        help='Reuse contexts between runs (makes reproduction impossible)')
    parser.add_argument(
        '--cc',
        dest='cc',
        type=str,
        default=CC,
        help="CC (default: $CC='{}')".format(CC))
    parser.add_argument(
        '--cxx',
        dest='cxx',
        type=str,
        default=CXX,
        help="CXX (default: $CXX='{}')".format(CXX))
    parser.add_argument(
        '--cppflags',
        dest='cppflags',
        type=str,
        default=CPPFLAGS,
        help="CPPFLAGS (default: $CPPFLAGS='{}')".format(CPPFLAGS))
    parser.add_argument(
        '--cflags',
        dest='cflags',
        type=str,
        default=CFLAGS,
        help="CFLAGS (default: $CFLAGS='{}')".format(CFLAGS))
    parser.add_argument(
        '--cxxflags',
        dest='cxxflags',
        type=str,
        default=CXXFLAGS,
        help="CXXFLAGS (default: $CXXFLAGS='{}')".format(CXXFLAGS))
    parser.add_argument(
        '--ldflags',
        dest='ldflags',
        type=str,
        default=LDFLAGS,
        help="LDFLAGS (default: $LDFLAGS='{}')".format(LDFLAGS))
    parser.add_argument(
        '--mflags',
        dest='mflags',
        type=str,
        default=MFLAGS,
        help="Extra Make flags (default: $MFLAGS='{}')".format(MFLAGS))
    parser.add_argument(
        'TARGET',
        nargs='*',
        type=str,
        help='Fuzz target(s) to build {{{}}}'.format(', '.join(ALL_TARGETS))
    )
    args = parser.parse_args(args)
    args = parse_env_flags(args, ' '.join(
        [args.cppflags, args.cflags, args.cxxflags, args.ldflags]))

    # Check option sanity
    if args.msan and (args.asan or args.ubsan):
        raise RuntimeError('MSAN may not be used with any other sanitizers')
    if args.msan_track_origins and not args.msan:
        raise RuntimeError('--enable-msan-track-origins requires MSAN')
    if args.ubsan_pointer_overflow and not args.ubsan:
        raise RuntimeError('--enable-ubsan-pointer-overflow requires UBSAN')
    if args.sanitize_recover and not args.sanitize:
        raise RuntimeError('--enable-sanitize-recover but no sanitizers used')

    return args


def build(args):
    try:
        args = build_parser(args)
    except Exception as e:
        print(e)
        return 1
    # The compilation flags we are setting
    targets = args.TARGET
    cc = args.cc
    cxx = args.cxx
    cppflags = shlex.split(args.cppflags)
    cflags = shlex.split(args.cflags)
    ldflags = shlex.split(args.ldflags)
    cxxflags = shlex.split(args.cxxflags)
    mflags = shlex.split(args.mflags)
    # Flags to be added to both cflags and cxxflags
    common_flags = []

    cppflags += [
        '-DDEBUGLEVEL={}'.format(args.debug),
        '-DMEM_FORCE_MEMORY_ACCESS={}'.format(args.memory_access),
        '-DFUZZ_RNG_SEED_SIZE={}'.format(args.fuzz_rng_seed_size),
    ]

    # Set flags for options
    assert not (args.fuzzer and args.coverage)
    if args.coverage:
        common_flags += [
            '-fsanitize-coverage=trace-pc-guard,indirect-calls,trace-cmp'
        ]
    if args.fuzzer:
        common_flags += ['-fsanitize=fuzzer']
        args.lib_fuzzing_engine = ''

    mflags += ['LIB_FUZZING_ENGINE={}'.format(args.lib_fuzzing_engine)]

    if args.sanitize_recover:
        recover_flags = ['-fsanitize-recover=all']
    else:
        recover_flags = ['-fno-sanitize-recover=all']
    if args.sanitize:
        common_flags += recover_flags

    if args.msan:
        msan_flags = ['-fsanitize=memory']
        if args.msan_track_origins:
            msan_flags += ['-fsanitize-memory-track-origins']
        common_flags += msan_flags
        # Append extra MSAN flags (it might require special setup)
        cppflags += [args.msan_extra_cppflags]
        cflags += [args.msan_extra_cflags]
        cxxflags += [args.msan_extra_cxxflags]
        ldflags += [args.msan_extra_ldflags]

    if args.asan:
        common_flags += ['-fsanitize=address']

    if args.ubsan:
        ubsan_flags = ['-fsanitize=undefined']
        if not args.ubsan_pointer_overflow:
            ubsan_flags += overflow_ubsan_flags(cc, cxx)
        common_flags += ubsan_flags

    if args.stateful_fuzzing:
        cppflags += ['-DSTATEFUL_FUZZING']

    if args.fuzzing_mode:
        cppflags += ['-DFUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION']

    if args.lib_fuzzing_engine == 'libregression.a':
        targets = ['libregression.a'] + targets

    # Append the common flags
    cflags += common_flags
    cxxflags += common_flags

    # Prepare the flags for Make
    cc_str = "CC={}".format(cc)
    cxx_str = "CXX={}".format(cxx)
    cppflags_str = "CPPFLAGS={}".format(' '.join(cppflags))
    cflags_str = "CFLAGS={}".format(' '.join(cflags))
    cxxflags_str = "CXXFLAGS={}".format(' '.join(cxxflags))
    ldflags_str = "LDFLAGS={}".format(' '.join(ldflags))

    # Print the flags
    print('MFLAGS={}'.format(' '.join(mflags)))
    print(cc_str)
    print(cxx_str)
    print(cppflags_str)
    print(cflags_str)
    print(cxxflags_str)
    print(ldflags_str)

    # Clean and build
    clean_cmd = ['make', 'clean'] + mflags
    print(' '.join(clean_cmd))
    subprocess.check_call(clean_cmd)
    build_cmd = [
        'make',
        cc_str,
        cxx_str,
        cppflags_str,
        cflags_str,
        cxxflags_str,
        ldflags_str,
    ] + mflags + targets
    print(' '.join(build_cmd))
    subprocess.check_call(build_cmd)
    return 0


def libfuzzer_parser(args):
    description = """
    Runs a libfuzzer binary.
    Passes all extra arguments to libfuzzer.
    The fuzzer should have been build with LIB_FUZZING_ENGINE pointing to
    libFuzzer.a.
    Generates output in the CORPORA directory, puts crashes in the ARTIFACT
    directory, and takes extra input from the SEED directory.
    To merge AFL's output pass the SEED as AFL's output directory and pass
    '-merge=1'.
    """
    parser = argparse.ArgumentParser(prog=args.pop(0), description=description)
    parser.add_argument(
        '--corpora',
        type=str,
        help='Override the default corpora dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET')))
    parser.add_argument(
        '--artifact',
        type=str,
        help='Override the default artifact dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET-crash')))
    parser.add_argument(
        '--seed',
        type=str,
        help='Override the default seed dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET-seed')))
    parser.add_argument(
        'TARGET',
        type=str,
        help='Fuzz target(s) to build {{{}}}'.format(', '.join(TARGETS)))
    args, extra = parser.parse_known_args(args)
    args.extra = extra

    if args.TARGET and args.TARGET not in TARGETS:
        raise RuntimeError('{} is not a valid target'.format(args.TARGET))

    return args


def libfuzzer(target, corpora=None, artifact=None, seed=None, extra_args=None):
    if corpora is None:
        corpora = abs_join(CORPORA_DIR, target)
    if artifact is None:
        artifact = abs_join(CORPORA_DIR, '{}-crash'.format(target))
    if seed is None:
        seed = abs_join(CORPORA_DIR, '{}-seed'.format(target))
    if extra_args is None:
        extra_args = []

    target = abs_join(FUZZ_DIR, target)

    corpora = [create(corpora)]
    artifact = create(artifact)
    seed = check(seed)

    corpora += [artifact]
    if seed is not None:
        corpora += [seed]

    cmd = [target, '-artifact_prefix={}/'.format(artifact)]
    cmd += corpora + extra_args
    print(' '.join(cmd))
    subprocess.check_call(cmd)


def libfuzzer_cmd(args):
    try:
        args = libfuzzer_parser(args)
    except Exception as e:
        print(e)
        return 1
    libfuzzer(args.TARGET, args.corpora, args.artifact, args.seed, args.extra)
    return 0


def afl_parser(args):
    description = """
    Runs an afl-fuzz job.
    Passes all extra arguments to afl-fuzz.
    The fuzzer should have been built with CC/CXX set to the AFL compilers,
    and with LIB_FUZZING_ENGINE='libregression.a'.
    Takes input from CORPORA and writes output to OUTPUT.
    Uses AFL_FUZZ as the binary (set from flag or environment variable).
    """
    parser = argparse.ArgumentParser(prog=args.pop(0), description=description)
    parser.add_argument(
        '--corpora',
        type=str,
        help='Override the default corpora dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET')))
    parser.add_argument(
        '--output',
        type=str,
        help='Override the default AFL output dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET-afl')))
    parser.add_argument(
        '--afl-fuzz',
        type=str,
        default=AFL_FUZZ,
        help='AFL_FUZZ (default: $AFL_FUZZ={})'.format(AFL_FUZZ))
    parser.add_argument(
        'TARGET',
        type=str,
        help='Fuzz target(s) to build {{{}}}'.format(', '.join(TARGETS)))
    args, extra = parser.parse_known_args(args)
    args.extra = extra

    if args.TARGET and args.TARGET not in TARGETS:
        raise RuntimeError('{} is not a valid target'.format(args.TARGET))

    if not args.corpora:
        args.corpora = abs_join(CORPORA_DIR, args.TARGET)
    if not args.output:
        args.output = abs_join(CORPORA_DIR, '{}-afl'.format(args.TARGET))

    return args


def afl(args):
    try:
        args = afl_parser(args)
    except Exception as e:
        print(e)
        return 1
    target = abs_join(FUZZ_DIR, args.TARGET)

    corpora = create(args.corpora)
    output = create(args.output)

    cmd = [args.afl_fuzz, '-i', corpora, '-o', output] + args.extra
    cmd += [target, '@@']
    print(' '.join(cmd))
    subprocess.call(cmd)
    return 0


def regression(args):
    try:
        description = """
        Runs one or more regression tests.
        The fuzzer should have been built with with
        LIB_FUZZING_ENGINE='libregression.a'.
        Takes input from CORPORA.
        """
        args = targets_parser(args, description)
    except Exception as e:
        print(e)
        return 1
    for target in args.TARGET:
        corpora = create(abs_join(CORPORA_DIR, target))
        target = abs_join(FUZZ_DIR, target)
        cmd = [target, corpora]
        print(' '.join(cmd))
        subprocess.check_call(cmd)
    return 0


def gen_parser(args):
    description = """
    Generate a seed corpus appropriate for TARGET with data generated with
    decodecorpus.
    The fuzz inputs are prepended with a seed before the zstd data, so the
    output of decodecorpus shouldn't be used directly.
    Generates NUMBER samples prepended with FUZZ_RNG_SEED_SIZE random bytes and
    puts the output in SEED.
    DECODECORPUS is the decodecorpus binary, and must already be built.
    """
    parser = argparse.ArgumentParser(prog=args.pop(0), description=description)
    parser.add_argument(
        '--number',
        '-n',
        type=int,
        default=100,
        help='Number of samples to generate')
    parser.add_argument(
        '--max-size-log',
        type=int,
        default=13,
        help='Maximum sample size to generate')
    parser.add_argument(
        '--seed',
        type=str,
        help='Override the default seed dir (default: {})'.format(
            abs_join(CORPORA_DIR, 'TARGET-seed')))
    parser.add_argument(
        '--decodecorpus',
        type=str,
        default=DECODECORPUS,
        help="decodecorpus binary (default: $DECODECORPUS='{}')".format(
            DECODECORPUS))
    parser.add_argument(
        '--fuzz-rng-seed-size',
        type=int,
        default=4,
        help="FUZZ_RNG_SEED_SIZE used for generate the samples (must match)"
    )
    parser.add_argument(
        'TARGET',
        type=str,
        help='Fuzz target(s) to build {{{}}}'.format(', '.join(TARGETS)))
    args, extra = parser.parse_known_args(args)
    args.extra = extra

    if args.TARGET and args.TARGET not in TARGETS:
        raise RuntimeError('{} is not a valid target'.format(args.TARGET))

    if not args.seed:
        args.seed = abs_join(CORPORA_DIR, '{}-seed'.format(args.TARGET))

    if not os.path.isfile(args.decodecorpus):
        raise RuntimeError("{} is not a file run 'make -C {} decodecorpus'".
                           format(args.decodecorpus, abs_join(FUZZ_DIR, '..')))

    return args


def gen(args):
    try:
        args = gen_parser(args)
    except Exception as e:
        print(e)
        return 1

    seed = create(args.seed)
    with tmpdir() as compressed:
        with tmpdir() as decompressed:
            cmd = [
                args.decodecorpus,
                '-n{}'.format(args.number),
                '-p{}/'.format(compressed),
                '-o{}'.format(decompressed),
            ]

            if 'block_' in args.TARGET:
                cmd += [
                    '--gen-blocks',
                    '--max-block-size-log={}'.format(args.max_size_log)
                ]
            else:
                cmd += ['--max-content-size-log={}'.format(args.max_size_log)]

            print(' '.join(cmd))
            subprocess.check_call(cmd)

            if '_round_trip' in args.TARGET:
                print('using decompressed data in {}'.format(decompressed))
                samples = decompressed
            elif '_decompress' in args.TARGET:
                print('using compressed data in {}'.format(compressed))
                samples = compressed

            # Copy the samples over and prepend the RNG seeds
            for name in os.listdir(samples):
                samplename = abs_join(samples, name)
                outname = abs_join(seed, name)
                rng_seed = os.urandom(args.fuzz_rng_seed_size)
                with open(samplename, 'rb') as sample:
                    with open(outname, 'wb') as out:
                        out.write(rng_seed)
                        CHUNK_SIZE = 131072
                        chunk = sample.read(CHUNK_SIZE)
                        while len(chunk) > 0:
                            out.write(chunk)
                            chunk = sample.read(CHUNK_SIZE)
    return 0


def minimize(args):
    try:
        description = """
        Runs a libfuzzer fuzzer with -merge=1 to build a minimal corpus in
        TARGET_seed_corpus. All extra args are passed to libfuzzer.
        """
        args = targets_parser(args, description)
    except Exception as e:
        print(e)
        return 1

    for target in args.TARGET:
        # Merge the corpus + anything else into the seed_corpus
        corpus = abs_join(CORPORA_DIR, target)
        seed_corpus = abs_join(CORPORA_DIR, "{}_seed_corpus".format(target))
        extra_args = [corpus, "-merge=1"] + args.extra
        libfuzzer(target, corpora=seed_corpus, extra_args=extra_args)
        seeds = set(os.listdir(seed_corpus))
        # Copy all crashes directly into the seed_corpus if not already present
        crashes = abs_join(CORPORA_DIR, '{}-crash'.format(target))
        for crash in os.listdir(crashes):
            if crash not in seeds:
                shutil.copy(abs_join(crashes, crash), seed_corpus)
                seeds.add(crash)


def zip_cmd(args):
    try:
        description = """
        Zips up the seed corpus.
        """
        args = targets_parser(args, description)
    except Exception as e:
        print(e)
        return 1

    for target in args.TARGET:
        # Zip the seed_corpus
        seed_corpus = abs_join(CORPORA_DIR, "{}_seed_corpus".format(target))
        zip_file = "{}.zip".format(seed_corpus)
        cmd = ["zip", "-r", "-q", "-j", "-9", zip_file, "."]
        print(' '.join(cmd))
        subprocess.check_call(cmd, cwd=seed_corpus)


def list_cmd(args):
    print("\n".join(TARGETS))


def short_help(args):
    name = args[0]
    print("Usage: {} [OPTIONS] COMMAND [ARGS]...\n".format(name))


def help(args):
    short_help(args)
    print("\tfuzzing helpers (select a command and pass -h for help)\n")
    print("Options:")
    print("\t-h, --help\tPrint this message")
    print("")
    print("Commands:")
    print("\tbuild\t\tBuild a fuzzer")
    print("\tlibfuzzer\tRun a libFuzzer fuzzer")
    print("\tafl\t\tRun an AFL fuzzer")
    print("\tregression\tRun a regression test")
    print("\tgen\t\tGenerate a seed corpus for a fuzzer")
    print("\tminimize\tMinimize the test corpora")
    print("\tzip\t\tZip the minimized corpora up")
    print("\tlist\t\tList the available targets")


def main():
    args = sys.argv
    if len(args) < 2:
        help(args)
        return 1
    if args[1] == '-h' or args[1] == '--help' or args[1] == '-H':
        help(args)
        return 1
    command = args.pop(1)
    args[0] = "{} {}".format(args[0], command)
    if command == "build":
        return build(args)
    if command == "libfuzzer":
        return libfuzzer_cmd(args)
    if command == "regression":
        return regression(args)
    if command == "afl":
        return afl(args)
    if command == "gen":
        return gen(args)
    if command == "minimize":
        return minimize(args)
    if command == "zip":
        return zip_cmd(args)
    if command == "list":
        return list_cmd(args)
    short_help(args)
    print("Error: No such command {} (pass -h for help)".format(command))
    return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2017-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import re
import sys
import sysconfig
import platform
import subprocess
from pathlib import Path

from setuptools import setup, Extension, find_namespace_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install_scripts import install_scripts
from distutils.version import LooseVersion
from setuptools.command.test import test as TestCommand


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.join(os.path.dirname(self.get_ext_fullpath(ext.name)), "tc2"))
        cmake_args = [
            # CMake sets the folders containing the libraries in the binaries RPATH upon build, so we can find them at run-time
            # Upon `install`, cmake clears the RPATH in the binaries, so they can use system-wide libraries (/usr/lib...)
            # However, if we do that, libraries installed with conda are lost (because they are not in system-wide directories)
            # Let's rather keep RPATH for all librairies outside of the build tree
            # See https://gitlab.kitware.com/cmake/community/-/wikis/doc/cmake/RPATH-handling#always-full-rpath
            '-DCMAKE_INSTALL_RPATH_USE_LINK_PATH=TRUE',
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
            '-DCMAKE_INSTALL_PREFIX=' + extdir,
            '-DPYTHON_EXECUTABLE=' + sys.executable,
            '-DPython3_ROOT_DIR=' + str(Path(sys.executable).parent.parent.absolute()),
            '-DPython3_FIND_STRATEGY=LOCATION'  # ... even if we find a newer version somewhere else in the system
        ]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += [
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(
                    cfg.upper(), extdir)
            ]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''), self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ['cmake', ext.sourcedir] + cmake_args,
            cwd=self.build_temp,
            env=env)
        subprocess.check_call(
            ['cmake', '--build', '.', '-j', '4'] + build_args, cwd=self.build_temp)
        subprocess.check_call(
            ['cmake', '--build', '.', '--target', 'install'],
            cwd=self.build_temp)


# Find version number in __init__
init_str = Path("tc2/__init__.py").read_text()
match = re.search(r"^__version__ = \"(?P<version>[\w\.]+?)\"", init_str, re.MULTILINE)
assert match is not None, "Could not find version in tc2/__init__.py"
version = match.group("version")


setup(
    name='tc2',
    version=version,
    author='Facebook AI Research',
    author_email='oncall+torchcraft@xmail.facebook.com',
    description='A fast and easy way to run StarCraft: Brood War as a Gym environment.',
    license='MIT',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    ext_modules=[CMakeExtension('tc2lib')],
    cmdclass=dict(build_ext=CMakeBuild),
    install_requires=(Path(__file__).parent / "requirements.txt").read_text().splitlines(),
    # And `apt-get install -y cmake libsdl2-dev libzmq3-dev` (debian)
    # or `brew install sdl2 zeromq` (OSX)
    zip_safe=False,
    packages=find_namespace_packages(include=['*tc2*'], exclude=['*test*']),
    package_data={"hiplot": ["py.typed"]},
    scripts=['scripts/tc2-setup'],
)



from collections import defaultdict
from tc2.constants import *
from typing import *


def eliminated_from_micro(tcstates, player_id: int) -> bool:
    """Elimination criterion for micro scenarios, where players have units but usually no buildings"""
    units = [len(tcstate.units[player_id]) for tcstate in tcstates]
    return max(units) == 0


def terminate_timeout(tcstates, seconds: int = 120) -> bool:
    return tcstates[TCAgent.client_id].frame_from_bwapi > 24 * seconds


def terminate_elimination_micro(tcstates):
    # Delay briefly to ensure that units have had a chance to spawn
    if tcstates[TCAgent.client_id].frame_from_bwapi < 3:
        return False
    agent_player_id = tcstates[TCAgent.client_id].player_id
    player_ids = [agent_player_id, other_player_id(agent_player_id)]
    return any(eliminated_from_micro(tcstates, player_id) for player_id in player_ids)


def reward_binary_micro(tcstates, player_id: int):
    if eliminated_from_micro(tcstates, player_id):
        return 0.0
    if eliminated_from_micro(tcstates, other_player_id(player_id)):
        return 1.0
    return 0.0


class TC2RewardAbstract:
    def __init__(self) -> None:
        pass

    def on_spawn(self, player_id: int, unit) -> None:
        pass

    def reward(self, tcstates) -> List[float]:
        raise Exception("The reward function is unimplemented")

    def done(self, tcstates) -> bool:
        raise Exception("The termination function is unimplemented")


class TC2RewardMicroBinary(TC2RewardAbstract):
    """A binary (0/1) reward function, which terminates after one minute or when a player has no remaining units."""

    def reward(self, tcstates) -> List[float]:
        agent_player_id = tcstates[TCAgent.client_id].player_id
        player_ids = [agent_player_id, other_player_id(agent_player_id)]
        return [reward_binary_micro(tcstates, player_id) for player_id in player_ids]

    def done(self, tcstates) -> bool:
        return terminate_timeout(tcstates) or terminate_elimination_micro(tcstates)


class TC2RewardMicroContinuousCount(TC2RewardMicroBinary):
    """A continuous reward function which measures how many units were defeated or retained."""

    def __init__(self) -> None:
        self._units_total: Dict[int, float] = defaultdict(float)
        super().__init__()

    def on_spawn(self, player_id: int, unit) -> None:
        if player_id not in self._units_total:
            self._units_total[player_id] = 0.0
        self._units_total[player_id] += 1.0

    def reward(self, tcstates) -> List[float]:
        """
        Reward = 50% binary reward, 25% units retained, 25% units defeated
        The reward is structured so that winning is always better than losing, no matter how many units were retained or defeated
        """
        def player_reward(player_id: int) -> float:
            opponent_id = other_player_id(player_id)
            units_retained = len(tcstates[player_id].units[player_id])
            units_defeated = len(tcstates[opponent_id].units[opponent_id])
            reward_binary = reward_binary_micro(tcstates, player_id)
            reward_retained = min(1.0, units_retained / max(1.0, self._units_total[player_id]))
            reward_defeated = 1.0 - min(1.0, units_defeated / max(1.0, self._units_total[opponent_id]))
            return 0.5 * reward_binary + 0.25 * reward_retained + 0.25 * reward_defeated
        agent_player_id = tcstates[TCAgent.client_id].player_id
        player_ids = [agent_player_id, other_player_id(agent_player_id)]
        return [player_reward(player_id) for player_id in player_ids]


class TC2RewardFullGameBinary(TC2RewardAbstract):
    """
    A binary (0/1) reward function, which terminates using the standard StarCraft termination condition:
    a player has no remaining buildings
    """

    def _is_alive(self, tcstates, player_id: int) -> bool:
        return any(TCStaticValues['isBuilding'][unit.type] for unit in tcstates[player_id].units[player_id])  # type: ignore

    def reward(self, tcstates) -> List[float]:
        agent_player_id = tcstates[TCAgent.client_id].player_id
        player_ids = [agent_player_id, other_player_id(agent_player_id)]
        alive: List[float] = [self._is_alive(tcstates, player_id) for player_id in player_ids]
        return [0.0 for player_id in player_ids] if all(alive) else alive

    def done(self, tcstates) -> bool:
        return max(self.reward(tcstates)) > 0


class TC2RewardResources(TC2RewardAbstract):
    def __init__(self, seconds: int = 180):
        self.seconds = seconds
        super().__init__()

    def reward(self, tcstates) -> List[float]:
        return [
            tcstate.frame.resources[tcstate.player_id].ore
            + tcstate.frame.resources[tcstate.player_id].gas
            for tcstate in tcstates]

    def done(self, tcstates) -> bool:
        return terminate_timeout(tcstates, seconds=self.seconds)


class TC2RewardUnitCount(TC2RewardAbstract):
    """A continuous reward function which measures how many (matching) units the player has at the end of the episode."""

    def __init__(self, seconds: int = 180, unit_criterion: Callable = lambda x: True):
        super().__init__()
        self.seconds = seconds
        self.unit_criterion = unit_criterion

    def reward(self, tcstates) -> List[float]:
        agent_player_id = tcstates[TCAgent.client_id].player_id
        return [len([unit for unit in tcstates[agent_player_id].units[agent_player_id] if self.unit_criterion(unit)]), 0]

    def done(self, tcstates) -> bool:
        return terminate_timeout(tcstates, seconds=self.seconds)


class TC2RewardWorkerCount(TC2RewardUnitCount):
    def __init__(self, seconds: int = 180):
        super().__init__(seconds=seconds, unit_criterion=(lambda unit: TCStaticValues['isWorker'][unit.type]))  # type: ignore


class TC2RewardMarineCount(TC2RewardUnitCount):
    def __init__(self, seconds: int = 180):
        super().__init__(seconds=seconds, unit_criterion=(lambda unit: unit.type == TCUnits.Terran_Marine))



#-----------------------------------------------------------------------------
#
#  Configuration for YouCompleteMe Vim plugin
#
#  http://valloric.github.io/YouCompleteMe/
#
#-----------------------------------------------------------------------------

from os.path import realpath, dirname

basedir = dirname(realpath(__file__))

# some default flags
# for more information install clang-3.2-doc package and
# check UsersManual.html
flags = [
'-Werror',
'-Wall',
'-Wextra',
'-pedantic',

'-std=c++11',

# '-x' and 'c++' also required
# use 'c' for C projects
'-x',
'c++',

# include third party libraries
'-I.',
]

# youcompleteme is calling this function to get flags
# You can also set database for flags. Check: JSONCompilationDatabase.html in
# clang-3.2-doc package
def FlagsForFile( filename ):
  return {
    'flags': flags,
    'do_cache': True
  }
